import ffmpeg from "ffmpeg";
import * as fffmpeg from "fluent-ffmpeg";
import { createCanvas, loadImage } from "canvas";
import fs from "fs";
import path from "path";
import * as dotenv from "dotenv";
import * as https from "https";
import * as http from "http";

dotenv.config();

const extractionFrameSize = Number(process.env.FRAME_SIZE) || 30;
const framesDirectory = process.env.FRAMES_DIR || "./frames";
const spritesDirectory = process.env.SPRITES_DIR || "./sprites";
const videosDirectory = process.env.VIDEOS_DIR || "./videos";
const spriteRowSize = Number(process.env.SPRITE_ROW_SIZE) || 6;
const spriteMaxLines = Number(process.env.SPRITE_MAX_LINES) || 12;
const sortAlphaNum = (a: string, b: string) =>
  a.localeCompare(b, "en", { numeric: true });

/**
 * Create sprite images and write to the filesystem based on input files, and needed width/height of the sprite
 *
 * @param files
 * @param width
 * @param height
 * @param videoDir
 */
const createSprites = async (
  files: string[],
  width: number,
  height: number,
  videoDir: string
): Promise<SpriteResult> => {
  const thumbWidth = Math.round(width / spriteRowSize); //e.g. 142
  const thumbHeight = Math.round(((width / spriteRowSize) * height) / width); //e.g. 80
  return new Promise<SpriteResult>(async (resolve, reject) => {
    try {
      const thumbnails: string[] = [];
      const canvas = createCanvas(
        thumbWidth * spriteRowSize,
        thumbHeight * spriteMaxLines
      );
      const ctx = canvas.getContext("2d");
      let currentPosition = 0;
      let currentLine = 0;
      let currentImg = 0;
      let dx = 0;
      let dy = 0;
      for (let i = 0; i <= files.length; i++) {
        const image = await loadImage(files[i]);
        console.log(
          `draw image[${i}] at dx[${dx}] dy[${dy}] w[${thumbWidth}] h[${thumbHeight}]`
        );
        ctx.drawImage(image, dx, dy, thumbWidth, thumbHeight);
        if (currentPosition < spriteRowSize - 1) {
          dx += thumbWidth;
          currentPosition += 1;
        } else {
          dx = 0;
          dy += thumbHeight;
          currentPosition = 0;
          currentLine += 1;
        }
        const spritePath = `${videoDir}/sprite-${currentImg}.jpeg`;
        if (currentLine === spriteMaxLines) {
          console.log("Sprite is full, write it to fs and reset canvas.");
          currentLine = 0;
          const buffer = canvas.toBuffer("image/jpeg");
          fs.writeFileSync(spritePath, buffer);
          thumbnails.push(spritePath);
          ctx.clearRect(0, 0, canvas.width, canvas.height);
          dx = 0;
          dy = 0;
          currentImg += 1;
        } else if (i + 1 === files.length) {
          console.log(
            "Last frame file is processed, write last non full image to fs."
          );
          const buffer = canvas.toBuffer("image/jpeg");
          fs.writeFileSync(spritePath, buffer);
          thumbnails.push(spritePath);
        }
        if (i + 1 === files.length) {
          resolve({
            sprites: thumbnails,
            thumbHeight,
            thumbWidth,
          });
        }
      }
    } catch (err) {
      reject(err);
    }
  });
};

/**
 * Checks if directory is accessible, if not creates a new one.
 *
 * @param videoDir
 * @param handler
 */
const checkDirAndWrite = (videoDir: string, handler: () => void) => {
  fs.access(videoDir, (error) => {
    if (error) {
      fs.mkdir(videoDir, (err) => {
        if (err) return;
        handler();
      });
    } else {
      cleanupDirectorySync(videoDir);
      handler();
    }
  });
};

/**
 * Delete if exist and recreate directories, and triggers callback afterwards
 *
 * @param directory
 * @param cb
 */
export const createDirectory = (directory: string, cb: () => void) => {
  fs.access(directory, (error) => {
    if (error) {
      fs.mkdir(directory, (err) => {
        if (err) return;
        cb();
      });
    } else cb();
  });
};

export const createDirectories = async (directories: string[]) => {
  for (const directory of directories) {
    try {
      await fs.accessSync(directory);
    } catch (err) {
      fs.mkdirSync(directory);
    }
  }
};

/**
 * Remove everything in the directory
 *
 * @param directory
 */
const cleanupDirectory = (directory: string) => {
  fs.readdir(directory, (err, files) => {
    if (err) console.error(err);
    for (const file of files) {
      fs.unlink(path.join(directory, file), (err) => {
        if (err) console.error(err);
      });
    }
  });
};

/**
 * Remove everything in the directory
 *
 * @param directory
 */
const cleanupDirectorySync = (directory: string) => {
  try {
    fs.readdirSync(directory).forEach((f) =>
      fs.unlinkSync(path.join(directory, f))
    );
  } catch (err) {
    console.log(err);
  }
};

/**
 * Remove downloaded video file after sprites generation
 *
 * @param src
 */
const deleteVideo = (src: string) => {
  fs.rm(src, (err) => {
    if (err) console.error(err);
  });
};

/**
 * Download file to local folder and trigger callback afterwards
 *
 * @param src
 * @param videoId
 * @param cb
 */
export const downloadVideoCb = (
  src: string,
  videoId: number,
  cb: (localSrc: string) => void
) => {
  createDirectory(videosDirectory, () => {
    const localPath = `${videosDirectory}/${videoId}.mp4`;
    console.log(`Download video to local path: ${localPath}`);
    const file = fs.createWriteStream(localPath);
    https.get(src, (response) => {
      response.pipe(file);
      console.log(`Download finished.`);
      cb(localPath);
    });
  });
};

export const downloadVideo: (url: string, videoId: string) => Promise<string> =
  async (url: string, videoId: string) => {
    const filePath = `${videosDirectory}/${videoId}.mp4`;

    console.log(`Download video to local path: ${filePath}`);

    const proto = !url.charAt(4).localeCompare("s") ? https : http;

    return new Promise((resolve, reject) => {
      const file = fs.createWriteStream(filePath);
      let fileInfo: any = null;

      const request = proto.get(url, (response) => {
        if (response.statusCode !== 200) {
          reject(new Error(`Failed to get '${url}' (${response.statusCode})`));
          return;
        }

        fileInfo = {
          path: filePath,
          mime: response.headers["content-type"],
          size: parseInt(response.headers["content-length"] as string, 10),
        };

        response.pipe(file);
      });

      // The destination stream is ended by the time it's called
      file.on("finish", () => resolve(filePath));

      request.on("error", (err) => {
        console.log(`Error when downloading video [${err}]`);
        fs.unlink(filePath, () => reject(err));
      });

      file.on("error", (err) => {
        console.log(`Error accessing file [${err}]`);
        fs.unlink(filePath, () => reject(err));
      });

      request.end();
    });
  };

export interface SpriteResult {
  sprites: string[];
  thumbWidth: number;
  thumbHeight: number;
}

const saveMetadata = (
  path: string,
  id: number,
  url: string,
  width: number,
  height: number
) => {
  fs.writeFileSync(
    `${path}/video-${id}.json`,
    JSON.stringify({
      id,
      url,
      thumbWidth: width,
      thumbHeight: height,
      spriteRowSize,
      spriteMaxLines,
    })
  );
};

/**
 * Extract frames from a video source.
 * Generate sprites based on that images.
 *
 * @param url
 * @param src
 * @param videoId
 */
export const extractFrames = async (
  url: string,
  src: string,
  videoId: number
) => {
  const videoFramesDir = `${framesDirectory}/${videoId}`;
  await createDirectories([
    videosDirectory,
    spritesDirectory,
    framesDirectory,
    videoFramesDir,
  ]);
  return new Promise<SpriteResult>(async (resolve, reject) => {
    try {
      new ffmpeg(src).then(
        async (video) => {
          video
            .fnExtractFrameToJPG(videoFramesDir, {
              every_n_frames: extractionFrameSize,
            })
            .then((files: string[]) => {
              const videoSpritesDir = `${spritesDirectory}/${videoId}`;
              checkDirAndWrite(videoSpritesDir, () =>
                fffmpeg.ffprobe(src, (err, data) => {
                  const { width, height } = data.streams.filter(
                    (s) => s.codec_type == "video"
                  )[0];
                  const widthOrDefault = width || 852;
                  const heightOrDefault = height || 480;
                  createSprites(
                    files.sort(sortAlphaNum),
                    widthOrDefault,
                    heightOrDefault,
                    videoSpritesDir
                  ).then((sprites) => {
                    console.log(
                      `Finish sprites generation [${sprites.sprites}].\nCleanup frames sub-directory [${videoFramesDir}] and video source [${src}] `
                    );
                    cleanupDirectory(videoFramesDir);
                    saveMetadata(
                      videoSpritesDir,
                      videoId,
                      url,
                      sprites.thumbWidth,
                      sprites.thumbHeight
                    );
                    deleteVideo(src);
                    resolve(sprites);
                  });
                })
              );
            })
            .catch((err) => {
              console.log(`Error: ${err}`);
              reject(err);
            });
        },
        (err: string) => {
          console.log(`Error: ${err}`);
          reject(err);
        }
      );
    } catch (e) {
      console.log(`Error: ${e.code} ${e.msg}`);
      reject(e);
    }
  });
};
